# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

from typing import TYPE_CHECKING
from typing import Any

from twisted.internet import defer

from buildbot.data import base
from buildbot.data import types
from buildbot.util.twisted import async_to_deferred

if TYPE_CHECKING:
    from buildbot.db.test_result_sets import TestResultSetModel
    from buildbot.util.twisted import InlineCallbacksType


def _db2data(model: TestResultSetModel):
    return {
        'test_result_setid': model.id,
        'builderid': model.builderid,
        'buildid': model.buildid,
        'stepid': model.stepid,
        'description': model.description,
        'category': model.category,
        'value_unit': model.value_unit,
        'tests_passed': model.tests_passed,
        'tests_failed': model.tests_failed,
        'complete': model.complete,
    }


class TestResultSetsEndpoint(base.BuildNestingMixin, base.Endpoint):
    kind = base.EndpointKind.COLLECTION
    pathPatterns = [
        "/test_result_sets",
        "/builders/n:builderid/test_result_sets",
        "/builders/s:buildername/test_result_sets",
        "/builds/n:buildid/test_result_sets",
        "/steps/n:stepid/test_result_sets",
    ]

    @defer.inlineCallbacks
    def get(self, resultSpec, kwargs):
        complete = resultSpec.popBooleanFilter('complete')
        if 'stepid' in kwargs:
            step_dbdict = yield self.master.db.steps.getStep(kwargs['stepid'])
            build_dbdict = yield self.master.db.builds.getBuild(step_dbdict.buildid)

            sets = yield self.master.db.test_result_sets.getTestResultSets(
                build_dbdict.builderid,
                buildid=step_dbdict.buildid,
                stepid=kwargs['stepid'],
                complete=complete,
                result_spec=resultSpec,
            )
        elif 'buildid' in kwargs:
            build_dbdict = yield self.master.db.builds.getBuild(kwargs['buildid'])

            sets = yield self.master.db.test_result_sets.getTestResultSets(
                build_dbdict.builderid,
                buildid=kwargs['buildid'],
                complete=complete,
                result_spec=resultSpec,
            )

        elif 'buildername' in kwargs or 'builderid' in kwargs:
            builderid = yield self.getBuilderId(kwargs)
            sets = yield self.master.db.test_result_sets.getTestResultSets(
                builderid, complete=complete, result_spec=resultSpec
            )
        else:
            sets = yield self.master.db.test_result_sets.getTestResultSets(
                complete=complete, result_spec=resultSpec
            )

        return [_db2data(model) for model in sets]


class TestResultSetsFromCommitRangeEndpoint(base.Endpoint):
    kind = base.EndpointKind.COLLECTION
    pathPatterns = [
        "/codebases/n:codebaseid/commit_range/n:commitid1/n:commitid2/test_result_sets",
    ]

    @async_to_deferred
    async def get(self, result_spec, kwargs) -> list[dict[str, Any]]:
        commit_from = int(kwargs.get('commitid1'))
        commit_to = int(kwargs.get('commitid2'))
        r = await self.master.db.codebase_commits.get_first_common_commit_with_ranges(
            commit_from, commit_to
        )
        if r is None:
            return []
        if r.to2_commit_ids[0] != commit_from:
            return []
        commit_ids = r.to2_commit_ids
        sets = await self.master.db.test_result_sets.get_test_result_sets_for_commits(
            commit_ids=commit_ids
        )
        return [_db2data(model) for model in sets]


class TestResultSetEndpoint(base.BuildNestingMixin, base.Endpoint):
    kind = base.EndpointKind.SINGLE
    pathPatterns = [
        "/test_result_sets/n:test_result_setid",
    ]

    @defer.inlineCallbacks
    def get(self, resultSpec, kwargs):
        model = yield self.master.db.test_result_sets.getTestResultSet(kwargs['test_result_setid'])
        return _db2data(model) if model else None


class TestResultSet(base.ResourceType):
    name = "test_result_set"
    plural = "test_result_sets"
    endpoints = [
        TestResultSetsEndpoint,
        TestResultSetsFromCommitRangeEndpoint,
        TestResultSetEndpoint,
    ]
    eventPathPatterns = [
        "/test_result_sets/:test_result_setid",
    ]

    class EntityType(types.Entity):
        test_result_setid = types.Integer()
        builderid = types.Integer()
        buildid = types.Integer()
        stepid = types.Integer()
        description = types.NoneOk(types.String())
        category = types.String()
        value_unit = types.String()
        tests_passed = types.NoneOk(types.Integer())
        tests_failed = types.NoneOk(types.Integer())
        complete = types.Boolean()

    entityType = EntityType(name)

    @defer.inlineCallbacks
    def generateEvent(self, test_result_setid, event):
        test_result_set = yield self.master.data.get(('test_result_sets', test_result_setid))
        self.produceEvent(test_result_set, event)

    @base.updateMethod
    @defer.inlineCallbacks
    def addTestResultSet(
        self,
        builderid: int,
        buildid: int,
        stepid: int,
        description: str,
        category: str,
        value_unit: str,
    ) -> InlineCallbacksType[int]:
        test_result_setid = yield self.master.db.test_result_sets.addTestResultSet(
            builderid, buildid, stepid, description, category, value_unit
        )
        yield self.generateEvent(test_result_setid, 'new')
        return test_result_setid

    @base.updateMethod
    @defer.inlineCallbacks
    def completeTestResultSet(
        self,
        test_result_setid: int,
        tests_passed: int | None = None,
        tests_failed: int | None = None,
    ) -> InlineCallbacksType[None]:
        yield self.master.db.test_result_sets.completeTestResultSet(
            test_result_setid, tests_passed, tests_failed
        )
        yield self.generateEvent(test_result_setid, 'completed')



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


__version__ = "4.0"


class Client:
    latest = None
    containerCreated = False
    start_exception = None

    def __init__(self, base_url):
        self.base_url = base_url
        self.call_args_create_container = []
        self.call_args_create_host_config = []
        self._images = [{'RepoTags': ['busybox:latest', 'worker:latest', 'tester:latest']}]
        self._pullable = ['alpine:latest', 'tester:latest']
        self._pullCount = 0
        self._containers = {}

        if Client.containerCreated:
            self.create_container("some-default-image")

    def images(self):
        return self._images

    def start(self, container):
        if self.start_exception is not None:
            raise self.start_exception  # pylint: disable=raising-bad-type

    def stop(self, id):
        pass

    def wait(self, id):
        return 0

    def build(self, fileobj, tag, pull, target):
        if fileobj.read() == b'BUG':
            pass
        elif pull != bool(pull):
            pass
        elif target != "":
            pass
        else:
            logs = []
            yield from logs
            self._images.append({'RepoTags': [tag + ':latest']})

    def pull(self, image, *args, **kwargs):
        if image in self._pullable:
            self._pullCount += 1
            self._images.append({'RepoTags': [image]})

    def containers(self, filters=None, *args, **kwargs):
        if filters is not None:
            if 'existing' in filters.get('name', ''):
                self.create_container(image='busybox:latest', name="buildbot-existing-87de7e")
                self.create_container(image='busybox:latest', name="buildbot-existing-87de7ef")

            return [c for c in self._containers.values() if c['name'].startswith(filters['name'])]
        return self._containers.values()

    def create_host_config(self, *args, **kwargs):
        self.call_args_create_host_config.append(kwargs)

    def create_container(self, image, *args, **kwargs):
        self.call_args_create_container.append(kwargs)
        name = kwargs.get('name', None)
        if 'buggy' in image:
            raise RuntimeError('we could not create this container')
        for c in self._containers.values():
            if c['name'] == name:
                raise RuntimeError('cannot create with same name')
        ret = {
            'Id': '8a61192da2b3bb2d922875585e29b74ec0dc4e0117fcbf84c962204e97564cd7',
            'Warnings': None,
        }
        self._containers[ret['Id']] = {
            'started': False,
            'image': image,
            'Id': ret['Id'],
            'name': name,  # docker does not return this
            'Names': ["/" + name],  # this what docker returns
            "State": "running",
        }
        return ret

    def remove_container(self, id, **kwargs):
        del self._containers[id]

    def logs(self, id, tail=None):
        return f"log for {id}\n1\n2\n3\nend\n".encode()

    def close(self):
        # dummy close, no connection to cleanup
        pass


class APIClient(Client):
    pass


class errors:
    class APIError(Exception):
        pass

    class NotFound(Exception):
        pass



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


import time

from twisted.internet import defer
from twisted.trial import unittest

from buildbot.db import steps
from buildbot.test import fakedb
from buildbot.test.fake import fakemaster
from buildbot.test.reactor import TestReactorMixin
from buildbot.util import epoch2datetime

TIME1 = 1304262222
TIME2 = 1304262223
TIME3 = 1304262224
TIME4 = 1304262235
TIME5 = 1304262236


class Tests(TestReactorMixin, unittest.TestCase):
    # common sample data

    backgroundData = [
        fakedb.Worker(id=47, name='linux'),
        fakedb.Buildset(id=20),
        fakedb.Builder(id=88, name='b1'),
        fakedb.BuildRequest(id=41, buildsetid=20, builderid=88),
        fakedb.Master(id=88),
        fakedb.Build(id=30, buildrequestid=41, number=7, masterid=88, builderid=88, workerid=47),
        fakedb.Build(id=31, buildrequestid=41, number=8, masterid=88, builderid=88, workerid=47),
    ]
    stepRows = [
        fakedb.Step(
            id=70,
            number=0,
            name='one',
            buildid=30,
            started_at=TIME1,
            locks_acquired_at=TIME2,
            complete_at=TIME3,
            state_string='test',
            results=0,
        ),
        fakedb.Step(
            id=71,
            number=1,
            name='two',
            buildid=30,
            started_at=TIME2,
            locks_acquired_at=TIME3,
            complete_at=TIME4,
            state_string='test',
            results=2,
            urls_json='[{"name": "url", "url": "http://url"}]',
            hidden=1,
        ),
        fakedb.Step(id=72, number=2, name='three', buildid=30, started_at=TIME5),
        fakedb.Step(id=73, number=0, name='wrong-build', buildid=31),
    ]
    stepDicts = [
        steps.StepModel(
            id=70,
            buildid=30,
            number=0,
            name='one',
            results=0,
            started_at=epoch2datetime(TIME1),
            locks_acquired_at=epoch2datetime(TIME2),
            complete_at=epoch2datetime(TIME3),
            state_string='test',
            urls=[],
            hidden=False,
        ),
        steps.StepModel(
            id=71,
            buildid=30,
            number=1,
            name='two',
            results=2,
            started_at=epoch2datetime(TIME2),
            locks_acquired_at=epoch2datetime(TIME3),
            complete_at=epoch2datetime(TIME4),
            state_string='test',
            urls=[steps.UrlModel(name='url', url='http://url')],
            hidden=True,
        ),
        steps.StepModel(
            id=72,
            buildid=30,
            number=2,
            name='three',
            results=None,
            started_at=epoch2datetime(TIME5),
            locks_acquired_at=None,
            complete_at=None,
            state_string='',
            urls=[],
            hidden=False,
        ),
    ]

    @defer.inlineCallbacks
    def setUp(self):
        self.setup_test_reactor()
        self.master = yield fakemaster.make_master(self, wantDb=True)
        self.db = self.master.db

    @defer.inlineCallbacks
    def test_getStep(self):
        yield self.db.insert_test_data([*self.backgroundData, self.stepRows[0]])
        stepdict = yield self.db.steps.getStep(70)
        self.assertIsInstance(stepdict, steps.StepModel)
        self.assertEqual(stepdict, self.stepDicts[0])

    @defer.inlineCallbacks
    def test_getStep_missing(self):
        stepdict = yield self.db.steps.getStep(50)
        self.assertEqual(stepdict, None)

    @defer.inlineCallbacks
    def test_getStep_number(self):
        yield self.db.insert_test_data([*self.backgroundData, self.stepRows[1]])
        stepdict = yield self.db.steps.getStep(buildid=30, number=1)
        self.assertIsInstance(stepdict, steps.StepModel)
        self.assertEqual(stepdict.id, 71)

    @defer.inlineCallbacks
    def test_getStep_number_missing(self):
        yield self.db.insert_test_data([*self.backgroundData, self.stepRows[1]])
        stepdict = yield self.db.steps.getStep(buildid=30, number=9)
        self.assertEqual(stepdict, None)

    @defer.inlineCallbacks
    def test_getStep_name(self):
        yield self.db.insert_test_data([*self.backgroundData, self.stepRows[2]])
        stepdict = yield self.db.steps.getStep(buildid=30, name='three')
        self.assertIsInstance(stepdict, steps.StepModel)
        self.assertEqual(stepdict.id, 72)

    @defer.inlineCallbacks
    def test_getStep_name_missing(self):
        yield self.db.insert_test_data([*self.backgroundData, self.stepRows[2]])
        stepdict = yield self.db.steps.getStep(buildid=30, name='five')
        self.assertEqual(stepdict, None)

    @defer.inlineCallbacks
    def test_getStep_invalid(self):
        with self.assertRaises(RuntimeError):
            yield self.db.steps.getStep(buildid=30)

    @defer.inlineCallbacks
    def test_getSteps(self):
        yield self.db.insert_test_data(self.backgroundData + self.stepRows)
        stepdicts = yield self.db.steps.getSteps(buildid=30)

        for stepdict in stepdicts:
            self.assertIsInstance(stepdict, steps.StepModel)

        self.assertEqual(stepdicts, self.stepDicts[:3])

    @defer.inlineCallbacks
    def test_getSteps_none(self):
        yield self.db.insert_test_data(self.backgroundData + self.stepRows)
        stepdicts = yield self.db.steps.getSteps(buildid=33)
        self.assertEqual(stepdicts, [])

    @defer.inlineCallbacks
    def test_addStep_getStep(self):
        yield self.db.insert_test_data(self.backgroundData)
        stepid, number, name = yield self.db.steps.addStep(
            buildid=30, name='new', state_string='new'
        )
        yield self.db.steps.startStep(stepid=stepid, started_at=TIME1, locks_acquired=False)
        self.assertEqual((number, name), (0, 'new'))
        stepdict = yield self.db.steps.getStep(stepid=stepid)
        self.assertIsInstance(stepdict, steps.StepModel)
        self.assertEqual(
            stepdict,
            steps.StepModel(
                id=stepid,
                buildid=30,
                name='new',
                number=0,
                started_at=epoch2datetime(TIME1),
                locks_acquired_at=None,
                complete_at=None,
                results=None,
                state_string='new',
                urls=[],
                hidden=False,
            ),
        )

    @defer.inlineCallbacks
    def test_addStep_getStep_locks_acquired_already(self):
        yield self.db.insert_test_data(self.backgroundData)
        stepid, number, name = yield self.db.steps.addStep(
            buildid=30, name='new', state_string='new'
        )
        yield self.db.steps.startStep(stepid=stepid, started_at=TIME1, locks_acquired=True)
        self.assertEqual((number, name), (0, 'new'))
        stepdict = yield self.db.steps.getStep(stepid=stepid)
        self.assertIsInstance(stepdict, steps.StepModel)
        self.assertEqual(
            stepdict,
            steps.StepModel(
                id=stepid,
                buildid=30,
                name="new",
                number=0,
                started_at=epoch2datetime(TIME1),
                locks_acquired_at=epoch2datetime(TIME1),
                complete_at=None,
                results=None,
                state_string="new",
                urls=[],
                hidden=False,
            ),
        )

    @defer.inlineCallbacks
    def test_addStep_getStep_locks_acquired_later(self):
        yield self.db.insert_test_data(self.backgroundData)
        stepid, number, name = yield self.db.steps.addStep(
            buildid=30, name='new', state_string='new'
        )
        yield self.db.steps.startStep(stepid=stepid, started_at=TIME1, locks_acquired=False)
        yield self.db.steps.set_step_locks_acquired_at(stepid=stepid, locks_acquired_at=TIME2)
        self.assertEqual((number, name), (0, 'new'))
        stepdict = yield self.db.steps.getStep(stepid=stepid)
        self.assertIsInstance(stepdict, steps.StepModel)
        self.assertEqual(
            stepdict,
            steps.StepModel(
                id=stepid,
                buildid=30,
                name='new',
                number=0,
                started_at=epoch2datetime(TIME1),
                locks_acquired_at=epoch2datetime(TIME2),
                complete_at=None,
                results=None,
                state_string='new',
                urls=[],
                hidden=False,
            ),
        )

    @defer.inlineCallbacks
    def test_addStep_getStep_existing_step(self):
        yield self.db.insert_test_data([*self.backgroundData, self.stepRows[0]])
        stepid, number, name = yield self.db.steps.addStep(
            buildid=30, name='new', state_string='new'
        )
        yield self.db.steps.startStep(stepid=stepid, started_at=TIME1, locks_acquired=False)
        self.assertEqual((number, name), (1, 'new'))
        stepdict = yield self.db.steps.getStep(stepid=stepid)
        self.assertIsInstance(stepdict, steps.StepModel)
        self.assertEqual(stepdict.number, number)
        self.assertEqual(stepdict.name, name)

    @defer.inlineCallbacks
    def test_addStep_getStep_name_collisions(self):
        yield self.db.insert_test_data([
            *self.backgroundData,
            fakedb.Step(id=73, number=0, name='new', buildid=30),
            fakedb.Step(id=74, number=1, name='new_1', buildid=30),
            fakedb.Step(id=75, number=2, name='new_2', buildid=30),
            fakedb.Step(id=76, number=3, name='new_step', buildid=30),
        ])
        stepid, number, name = yield self.db.steps.addStep(
            buildid=30, name='new', state_string='new'
        )
        yield self.db.steps.startStep(stepid=stepid, started_at=TIME1, locks_acquired=False)
        self.assertEqual((number, name), (4, 'new_3'))
        stepdict = yield self.db.steps.getStep(stepid=stepid)
        self.assertIsInstance(stepdict, steps.StepModel)
        self.assertEqual(stepdict.number, number)
        self.assertEqual(stepdict.name, name)

    @defer.inlineCallbacks
    def test_setStepStateString(self):
        yield self.db.insert_test_data([*self.backgroundData, self.stepRows[2]])
        yield self.db.steps.setStepStateString(stepid=72, state_string='aaa')
        stepdict = yield self.db.steps.getStep(stepid=72)
        self.assertEqual(stepdict.state_string, 'aaa')

    @defer.inlineCallbacks
    def test_addURL(self):
        yield self.db.insert_test_data([*self.backgroundData, self.stepRows[2]])
        yield self.db.steps.addURL(stepid=72, name='foo', url='bar')

        stepdict = yield self.db.steps.getStep(stepid=72)
        self.assertEqual(stepdict.urls, [steps.UrlModel(name='foo', url='bar')])

    @defer.inlineCallbacks
    def test_addURL_race(self):
        yield self.db.insert_test_data([*self.backgroundData, self.stepRows[2]])
        yield defer.gatherResults(
            [
                # only a tiny sleep is required to see the problem.
                self.db.steps.addURL(
                    stepid=72, name='foo', url='bar', _racehook=lambda: time.sleep(0.01)
                ),
                self.db.steps.addURL(stepid=72, name='foo2', url='bar2'),
            ],
            consumeErrors=True,
        )

        stepdict = yield self.db.steps.getStep(stepid=72)

        def urlKey(url):
            return url.name

        # order is not guaranteed though
        self.assertEqual(
            sorted(stepdict.urls, key=urlKey),
            sorted(
                [steps.UrlModel(name='foo', url='bar'), steps.UrlModel(name='foo2', url='bar2')],
                key=urlKey,
            ),
        )

    @defer.inlineCallbacks
    def test_addURL_no_duplicate(self):
        yield self.db.insert_test_data([*self.backgroundData, self.stepRows[2]])
        yield defer.gatherResults(
            [
                self.db.steps.addURL(stepid=72, name='foo', url='bar'),
                self.db.steps.addURL(stepid=72, name='foo', url='bar'),
            ],
            consumeErrors=True,
        )

        stepdict = yield self.db.steps.getStep(stepid=72)

        self.assertEqual(stepdict.urls, [steps.UrlModel(name='foo', url='bar')])

    @defer.inlineCallbacks
    def test_finishStep(self):
        self.reactor.advance(TIME2)
        yield self.db.insert_test_data([*self.backgroundData, self.stepRows[2]])
        yield self.db.steps.finishStep(stepid=72, results=11, hidden=False)
        stepdict = yield self.db.steps.getStep(stepid=72)
        self.assertEqual(stepdict.results, 11)
        self.assertEqual(stepdict.complete_at, epoch2datetime(TIME2))
        self.assertEqual(stepdict.hidden, False)

    @defer.inlineCallbacks
    def test_finishStep_hidden(self):
        yield self.db.insert_test_data([*self.backgroundData, self.stepRows[2]])
        yield self.db.steps.finishStep(stepid=72, results=11, hidden=True)
        stepdict = yield self.db.steps.getStep(stepid=72)
        self.assertEqual(stepdict.hidden, True)

    @defer.inlineCallbacks
    def test_addStep_getStep_name_collisions_too_long(self):
        yield self.db.insert_test_data([
            *self.backgroundData,
            fakedb.Step(id=73, number=0, name='a' * 49, buildid=30),
            fakedb.Step(id=74, number=1, name='a' * 48 + '_1', buildid=30),
        ])
        stepid, number, name = yield self.db.steps.addStep(
            buildid=30, name='a' * 49, state_string='new'
        )
        yield self.db.steps.startStep(stepid=stepid, started_at=TIME1, locks_acquired=False)
        self.assertEqual((number, name), (2, 'a' * 48 + '_2'))
        stepdict = yield self.db.steps.getStep(stepid=stepid)
        self.assertIsInstance(stepdict, steps.StepModel)
        self.assertEqual(stepdict.number, number)
        self.assertEqual(stepdict.name, name)

    @defer.inlineCallbacks
    def test_addStep_getStep_name_collisions_too_long_extra_digits(self):
        yield self.db.insert_test_data(
            self.backgroundData
            + [
                fakedb.Step(id=73, number=0, name='a' * 50, buildid=30),
            ]
            + [
                fakedb.Step(id=73 + i, number=i, name='a' * 48 + (f'_{i}'), buildid=30)
                for i in range(1, 10)
            ]
            + [
                fakedb.Step(id=73 + i, number=i, name='a' * 47 + (f'_{i}'), buildid=30)
                for i in range(10, 100)
            ]
        )
        stepid, number, name = yield self.db.steps.addStep(
            buildid=30, name='a' * 50, state_string='new'
        )
        yield self.db.steps.startStep(stepid=stepid, started_at=TIME1, locks_acquired=False)
        self.assertEqual((number, name), (100, 'a' * 46 + '_100'))
        stepdict = yield self.db.steps.getStep(stepid=stepid)
        self.assertIsInstance(stepdict, steps.StepModel)
        self.assertEqual(stepdict.number, number)
        self.assertEqual(stepdict.name, name)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from twisted.internet import defer
from twisted.trial import unittest

from buildbot.data import connector
from buildbot.data import root
from buildbot.test.util import endpoint
from buildbot.test.util.warnings import assertProducesWarnings
from buildbot.warnings import DeprecatedApiWarning


class RootEndpoint(endpoint.EndpointMixin, unittest.TestCase):
    endpointClass = root.RootEndpoint
    resourceTypeClass = root.Root

    @defer.inlineCallbacks
    def setUp(self):
        yield self.setUpEndpoint()
        self.master.data.rootLinks = [
            {'name': 'abc'},
        ]

    @defer.inlineCallbacks
    def test_get(self):
        with assertProducesWarnings(
            DeprecatedApiWarning, message_pattern='.*the root endpoint with endpoint directory.*'
        ):
            rootlinks = yield self.callGet(('',))

        for rootlink in rootlinks:
            self.validateData(rootlink)

        self.assertEqual(
            rootlinks,
            [
                {'name': 'abc'},
            ],
        )


class SpecEndpoint(endpoint.EndpointMixin, unittest.TestCase):
    endpointClass = root.SpecEndpoint
    resourceTypeClass = root.Spec

    @defer.inlineCallbacks
    def setUp(self):
        yield self.setUpEndpoint()
        # replace fakeConnector with real DataConnector
        self.master.data.disownServiceParent()
        self.master.data = connector.DataConnector()
        yield self.master.data.setServiceParent(self.master)

    @defer.inlineCallbacks
    def test_get(self):
        specs = yield self.callGet(('application.spec',))

        for s in specs:
            self.validateData(s)

        for s in specs:
            # only test an endpoint that is reasonably stable
            if s['path'] != "master":
                continue
            self.assertEqual(
                s,
                {
                    'path': 'master',
                    'type': 'master',
                    'type_spec': {
                        'fields': [
                            {'name': 'active', 'type': 'boolean', 'type_spec': {'name': 'boolean'}},
                            {
                                'name': 'masterid',
                                'type': 'integer',
                                'type_spec': {'name': 'integer'},
                            },
                            {'name': 'link', 'type': 'link', 'type_spec': {'name': 'link'}},
                            {'name': 'name', 'type': 'string', 'type_spec': {'name': 'string'}},
                            {
                                'name': 'last_active',
                                'type': 'datetime',
                                'type_spec': {'name': 'datetime'},
                            },
                        ],
                        'type': 'master',
                    },
                    'plural': 'masters',
                },
            )



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

import sqlalchemy as sa
from sqlalchemy.inspection import inspect
from twisted.trial import unittest

from buildbot.test.util import migration
from buildbot.util import sautils


class Migration(migration.MigrateTestMixin, unittest.TestCase):
    def setUp(self):
        return self.setUpMigrateTest()

    def create_tables_thd(self, conn):
        metadata = sa.MetaData()
        metadata.bind = conn

        # parent_buildid foreign key is removed for the purposes of the test
        buildsets = sautils.Table(
            "buildsets",
            metadata,
            sa.Column("id", sa.Integer, primary_key=True),
            sa.Column("external_idstring", sa.String(256)),
            sa.Column("reason", sa.String(256)),
            sa.Column("submitted_at", sa.Integer, nullable=False),
            sa.Column(
                "complete", sa.SmallInteger, nullable=False, server_default=sa.DefaultClause("0")
            ),
            sa.Column("complete_at", sa.Integer),
            sa.Column("results", sa.SmallInteger),
            sa.Column("parent_relationship", sa.Text),
        )
        buildsets.create(bind=conn)

        conn.execute(
            buildsets.insert(),
            [
                {
                    "id": 4,
                    "external_idstring": 5,
                    "reason": "rebuild",
                    "submitted_at": 1695730972,
                    "complete": 1,
                    "complete_at": 1695730977,
                    "results": 0,
                    "parent_relationship": "Triggered from",
                }
            ],
        )
        conn.commit()

        builds = sautils.Table("builds", metadata, sa.Column("id", sa.Integer, primary_key=True))
        builds.create(bind=conn)

        conn.execute(
            builds.insert(),
            [
                {
                    "id": 123,
                }
            ],
        )
        conn.commit()

    def test_update(self):
        def setup_thd(conn):
            self.create_tables_thd(conn)

        def verify_thd(conn):
            metadata = sa.MetaData()
            metadata.bind = conn

            # check that builsets.rebuilt_buildid has been added
            buildsets = sautils.Table('buildsets', metadata, autoload_with=conn)
            self.assertIsInstance(buildsets.c.rebuilt_buildid.type, sa.Integer)

            q = sa.select(
                buildsets.c.rebuilt_buildid,
            )

            all_fk_info = inspect(conn).get_foreign_keys("buildsets")
            fk_in_search = []
            for fk in all_fk_info:
                if fk["name"] == "rebuilt_buildid":
                    fk_in_search.append(fk)
                # verify that a foreign with name "fk_buildsets_rebuilt_buildid" was found
                self.assertEqual(len(fk_in_search), 1)

            conn.execute(
                buildsets.insert(),
                [
                    {
                        "id": 5,
                        "external_idstring": 6,
                        "reason": "rebuild",
                        "submitted_at": 1695730973,
                        "complete": 1,
                        "complete_at": 1695730978,
                        "results": 0,
                        "rebuilt_buildid": 123,
                        "parent_relationship": "Triggered from",
                    }
                ],
            )
            conn.commit()

            rebuilt_buildid_list = []
            for row in conn.execute(q):
                rebuilt_buildid_list.append(row.rebuilt_buildid)

            # verify that the 1st default value was set correctly to None
            self.assertEqual(rebuilt_buildid_list, [None, 123])

        return self.do_test_migration('064', '065', setup_thd, verify_thd)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from copy import deepcopy
from unittest import mock

from twisted.internet import defer
from twisted.trial import unittest
from zope.interface import implementer

from buildbot.interfaces import IRenderable
from buildbot.process.buildrequest import TempChange
from buildbot.process.buildrequest import TempSourceStamp
from buildbot.process.properties import FlattenList
from buildbot.process.properties import Interpolate
from buildbot.process.properties import Properties
from buildbot.process.properties import PropertiesMixin
from buildbot.process.properties import Property
from buildbot.process.properties import Transform
from buildbot.process.properties import WithProperties
from buildbot.process.properties import _Lazy
from buildbot.process.properties import _Lookup
from buildbot.process.properties import _SourceStampDict
from buildbot.process.properties import renderer
from buildbot.test.fake.fakebuild import FakeBuild
from buildbot.test.util.config import ConfigErrorsMixin
from buildbot.test.util.properties import ConstantRenderable


class FakeSource:
    def __init__(self):
        self.branch = None
        self.codebase = ''
        self.project = ''
        self.repository = ''
        self.revision = None

    def asDict(self):
        ds = {
            'branch': self.branch,
            'codebase': self.codebase,
            'project': self.project,
            'repository': self.repository,
            'revision': self.revision,
        }
        return ds


@implementer(IRenderable)
class DeferredRenderable:
    def __init__(self):
        self.d = defer.Deferred()

    def getRenderingFor(self, build):
        return self.d

    def callback(self, value):
        self.d.callback(value)


class TestPropertyMap(unittest.TestCase):
    """
    Test the behavior of PropertyMap, using the external interface
    provided by WithProperties.
    """

    def setUp(self):
        self.props = Properties(
            prop_str='a-string',
            prop_none=None,
            prop_list=['a', 'b'],
            prop_zero=0,
            prop_one=1,
            prop_false=False,
            prop_true=True,
            prop_empty='',
        )
        self.build = FakeBuild(props=self.props)

    @defer.inlineCallbacks
    def doTestSimpleWithProperties(self, fmtstring, expect, **kwargs):
        res = yield self.build.render(WithProperties(fmtstring, **kwargs))
        self.assertEqual(res, f"{expect}")

    def testSimpleStr(self):
        return self.doTestSimpleWithProperties('%(prop_str)s', 'a-string')

    def testSimpleNone(self):
        # None is special-cased to become an empty string
        return self.doTestSimpleWithProperties('%(prop_none)s', '')

    def testSimpleList(self):
        return self.doTestSimpleWithProperties('%(prop_list)s', ['a', 'b'])

    def testSimpleZero(self):
        return self.doTestSimpleWithProperties('%(prop_zero)s', 0)

    def testSimpleOne(self):
        return self.doTestSimpleWithProperties('%(prop_one)s', 1)

    def testSimpleFalse(self):
        return self.doTestSimpleWithProperties('%(prop_false)s', False)

    def testSimpleTrue(self):
        return self.doTestSimpleWithProperties('%(prop_true)s', True)

    def testSimpleEmpty(self):
        return self.doTestSimpleWithProperties('%(prop_empty)s', '')

    @defer.inlineCallbacks
    def testSimpleUnset(self):
        with self.assertRaises(KeyError):
            yield self.build.render(WithProperties('%(prop_nosuch)s'))

    def testColonMinusSet(self):
        return self.doTestSimpleWithProperties('%(prop_str:-missing)s', 'a-string')

    def testColonMinusNone(self):
        # None is special-cased here, too
        return self.doTestSimpleWithProperties('%(prop_none:-missing)s', '')

    def testColonMinusZero(self):
        return self.doTestSimpleWithProperties('%(prop_zero:-missing)s', 0)

    def testColonMinusOne(self):
        return self.doTestSimpleWithProperties('%(prop_one:-missing)s', 1)

    def testColonMinusFalse(self):
        return self.doTestSimpleWithProperties('%(prop_false:-missing)s', False)

    def testColonMinusTrue(self):
        return self.doTestSimpleWithProperties('%(prop_true:-missing)s', True)

    def testColonMinusEmpty(self):
        return self.doTestSimpleWithProperties('%(prop_empty:-missing)s', '')

    def testColonMinusUnset(self):
        return self.doTestSimpleWithProperties('%(prop_nosuch:-missing)s', 'missing')

    def testColonTildeSet(self):
        return self.doTestSimpleWithProperties('%(prop_str:~missing)s', 'a-string')

    def testColonTildeNone(self):
        # None is special-cased *differently* for ~:
        return self.doTestSimpleWithProperties('%(prop_none:~missing)s', 'missing')

    def testColonTildeZero(self):
        return self.doTestSimpleWithProperties('%(prop_zero:~missing)s', 'missing')

    def testColonTildeOne(self):
        return self.doTestSimpleWithProperties('%(prop_one:~missing)s', 1)

    def testColonTildeFalse(self):
        return self.doTestSimpleWithProperties('%(prop_false:~missing)s', 'missing')

    def testColonTildeTrue(self):
        return self.doTestSimpleWithProperties('%(prop_true:~missing)s', True)

    def testColonTildeEmpty(self):
        return self.doTestSimpleWithProperties('%(prop_empty:~missing)s', 'missing')

    def testColonTildeUnset(self):
        return self.doTestSimpleWithProperties('%(prop_nosuch:~missing)s', 'missing')

    def testColonPlusSet(self):
        return self.doTestSimpleWithProperties('%(prop_str:+present)s', 'present')

    def testColonPlusNone(self):
        return self.doTestSimpleWithProperties('%(prop_none:+present)s', 'present')

    def testColonPlusZero(self):
        return self.doTestSimpleWithProperties('%(prop_zero:+present)s', 'present')

    def testColonPlusOne(self):
        return self.doTestSimpleWithProperties('%(prop_one:+present)s', 'present')

    def testColonPlusFalse(self):
        return self.doTestSimpleWithProperties('%(prop_false:+present)s', 'present')

    def testColonPlusTrue(self):
        return self.doTestSimpleWithProperties('%(prop_true:+present)s', 'present')

    def testColonPlusEmpty(self):
        return self.doTestSimpleWithProperties('%(prop_empty:+present)s', 'present')

    def testColonPlusUnset(self):
        return self.doTestSimpleWithProperties('%(prop_nosuch:+present)s', '')

    @defer.inlineCallbacks
    def testClearTempValues(self):
        yield self.doTestSimpleWithProperties('', '', prop_temp=lambda b: 'present')
        yield self.doTestSimpleWithProperties('%(prop_temp:+present)s', '')

    def testTempValue(self):
        return self.doTestSimpleWithProperties(
            '%(prop_temp)s', 'present', prop_temp=lambda b: 'present'
        )

    def testTempValueOverrides(self):
        return self.doTestSimpleWithProperties('%(prop_one)s', 2, prop_one=lambda b: 2)

    def testTempValueColonMinusSet(self):
        return self.doTestSimpleWithProperties('%(prop_one:-missing)s', 2, prop_one=lambda b: 2)

    def testTempValueColonMinusUnset(self):
        return self.doTestSimpleWithProperties(
            '%(prop_nosuch:-missing)s', 'temp', prop_nosuch=lambda b: 'temp'
        )

    def testTempValueColonTildeTrueSet(self):
        return self.doTestSimpleWithProperties(
            '%(prop_false:~nontrue)s', 'temp', prop_false=lambda b: 'temp'
        )

    def testTempValueColonTildeTrueUnset(self):
        return self.doTestSimpleWithProperties(
            '%(prop_nosuch:~nontrue)s', 'temp', prop_nosuch=lambda b: 'temp'
        )

    def testTempValueColonTildeFalseFalse(self):
        return self.doTestSimpleWithProperties(
            '%(prop_false:~nontrue)s', 'nontrue', prop_false=lambda b: False
        )

    def testTempValueColonTildeTrueFalse(self):
        return self.doTestSimpleWithProperties(
            '%(prop_true:~nontrue)s', True, prop_true=lambda b: False
        )

    def testTempValueColonTildeNoneFalse(self):
        return self.doTestSimpleWithProperties(
            '%(prop_nosuch:~nontrue)s', 'nontrue', prop_nosuch=lambda b: False
        )

    def testTempValueColonTildeFalseZero(self):
        return self.doTestSimpleWithProperties(
            '%(prop_false:~nontrue)s', 'nontrue', prop_false=lambda b: 0
        )

    def testTempValueColonTildeTrueZero(self):
        return self.doTestSimpleWithProperties(
            '%(prop_true:~nontrue)s', True, prop_true=lambda b: 0
        )

    def testTempValueColonTildeNoneZero(self):
        return self.doTestSimpleWithProperties(
            '%(prop_nosuch:~nontrue)s', 'nontrue', prop_nosuch=lambda b: 0
        )

    def testTempValueColonTildeFalseBlank(self):
        return self.doTestSimpleWithProperties(
            '%(prop_false:~nontrue)s', 'nontrue', prop_false=lambda b: ''
        )

    def testTempValueColonTildeTrueBlank(self):
        return self.doTestSimpleWithProperties(
            '%(prop_true:~nontrue)s', True, prop_true=lambda b: ''
        )

    def testTempValueColonTildeNoneBlank(self):
        return self.doTestSimpleWithProperties(
            '%(prop_nosuch:~nontrue)s', 'nontrue', prop_nosuch=lambda b: ''
        )

    def testTempValuePlusSetSet(self):
        return self.doTestSimpleWithProperties('%(prop_one:+set)s', 'set', prop_one=lambda b: 2)

    def testTempValuePlusUnsetSet(self):
        return self.doTestSimpleWithProperties(
            '%(prop_nosuch:+set)s', 'set', prop_nosuch=lambda b: 1
        )


class TestInterpolateConfigure(unittest.TestCase, ConfigErrorsMixin):
    """
    Test that Interpolate reports errors in the interpolation string
    at configure time.
    """

    def test_invalid_args_and_kwargs(self):
        with self.assertRaisesConfigError("Interpolate takes either positional"):
            Interpolate("%s %(foo)s", 1, foo=2)

    def test_invalid_selector(self):
        with self.assertRaisesConfigError("invalid Interpolate selector 'garbage'"):
            Interpolate("%(garbage:test)s")

    def test_no_selector(self):
        with self.assertRaisesConfigError(
            "invalid Interpolate substitution without selector 'garbage'"
        ):
            Interpolate("%(garbage)s")

    def test_invalid_default_type(self):
        with self.assertRaisesConfigError("invalid Interpolate default type '@'"):
            Interpolate("%(prop:some_prop:@wacky)s")

    def test_nested_invalid_selector(self):
        with self.assertRaisesConfigError("invalid Interpolate selector 'garbage'"):
            Interpolate("%(prop:some_prop:~%(garbage:test)s)s")

    def test_colon_ternary_missing_delimeter(self):
        with self.assertRaisesConfigError(
            "invalid Interpolate ternary expression 'one' with delimiter ':'"
        ):
            Interpolate("echo '%(prop:P:?:one)s'")

    def test_colon_ternary_paren_delimiter(self):
        with self.assertRaisesConfigError(
            "invalid Interpolate ternary expression 'one(:)' with delimiter ':'"
        ):
            Interpolate("echo '%(prop:P:?:one(:))s'")

    def test_colon_ternary_hash_bad_delimeter(self):
        with self.assertRaisesConfigError(
            "invalid Interpolate ternary expression 'one' with delimiter '|'"
        ):
            Interpolate("echo '%(prop:P:#?|one)s'")

    def test_prop_invalid_character(self):
        with self.assertRaisesConfigError(
            "Property name must be alphanumeric for prop Interpolation 'a+a'"
        ):
            Interpolate("echo '%(prop:a+a)s'")

    def test_kw_invalid_character(self):
        with self.assertRaisesConfigError(
            "Keyword must be alphanumeric for kw Interpolation 'a+a'"
        ):
            Interpolate("echo '%(kw:a+a)s'")

    def test_src_codebase_invalid_character(self):
        with self.assertRaisesConfigError(
            "Codebase must be alphanumeric for src Interpolation 'a+a:a'"
        ):
            Interpolate("echo '%(src:a+a:a)s'")

    def test_src_attr_invalid_character(self):
        with self.assertRaisesConfigError(
            "Attribute must be alphanumeric for src Interpolation 'a:a+a'"
        ):
            Interpolate("echo '%(src:a:a+a)s'")

    def test_src_missing_attr(self):
        with self.assertRaisesConfigError("Must specify both codebase and attr"):
            Interpolate("echo '%(src:a)s'")


class TestInterpolatePositional(unittest.TestCase):
    def setUp(self):
        self.props = Properties()
        self.build = FakeBuild(props=self.props)

    @defer.inlineCallbacks
    def test_string(self):
        command = Interpolate("test %s", "one fish")
        rendered = yield self.build.render(command)
        self.assertEqual(rendered, "test one fish")

    @defer.inlineCallbacks
    def test_twoString(self):
        command = Interpolate("test %s, %s", "one fish", "two fish")
        rendered = yield self.build.render(command)
        self.assertEqual(rendered, "test one fish, two fish")

    def test_deferred(self):
        renderable = DeferredRenderable()
        command = Interpolate("echo '%s'", renderable)
        d = self.build.render(command)
        d.addCallback(self.assertEqual, "echo 'red fish'")
        renderable.callback("red fish")
        return d

    @defer.inlineCallbacks
    def test_renderable(self):
        self.props.setProperty("buildername", "blue fish", "test")
        command = Interpolate("echo '%s'", Property("buildername"))
        rendered = yield self.build.render(command)
        self.assertEqual(rendered, "echo 'blue fish'")


class TestInterpolateProperties(unittest.TestCase):
    def setUp(self):
        self.props = Properties()
        self.build = FakeBuild(props=self.props)

    @defer.inlineCallbacks
    def test_properties(self):
        self.props.setProperty("buildername", "winbld", "test")
        command = Interpolate("echo buildby-%(prop:buildername)s")
        rendered = yield self.build.render(command)
        self.assertEqual(rendered, "echo buildby-winbld")

    @defer.inlineCallbacks
    def test_properties_newline(self):
        self.props.setProperty("buildername", "winbld", "test")
        command = Interpolate("aa\n%(prop:buildername)s\nbb")
        rendered = yield self.build.render(command)
        self.assertEqual(rendered, "aa\nwinbld\nbb")

    @defer.inlineCallbacks
    def test_property_not_set(self):
        command = Interpolate("echo buildby-%(prop:buildername)s")
        rendered = yield self.build.render(command)
        self.assertEqual(rendered, "echo buildby-")

    @defer.inlineCallbacks
    def test_property_colon_minus(self):
        command = Interpolate("echo buildby-%(prop:buildername:-blddef)s")
        rendered = yield self.build.render(command)
        self.assertEqual(rendered, "echo buildby-blddef")

    @defer.inlineCallbacks
    def test_deepcopy(self):
        # After a deepcopy, Interpolate instances used to lose track
        # that they didn't have a ``hasKey`` value
        # see http://trac.buildbot.net/ticket/3505
        self.props.setProperty("buildername", "linux4", "test")
        command = deepcopy(Interpolate("echo buildby-%(prop:buildername:-blddef)s"))
        rendered = yield self.build.render(command)
        self.assertEqual(rendered, "echo buildby-linux4")

    @defer.inlineCallbacks
    def test_property_colon_tilde_true(self):
        self.props.setProperty("buildername", "winbld", "test")
        command = Interpolate("echo buildby-%(prop:buildername:~blddef)s")
        rendered = yield self.build.render(command)
        self.assertEqual(rendered, "echo buildby-winbld")

    @defer.inlineCallbacks
    def test_property_colon_tilde_false(self):
        self.props.setProperty("buildername", "", "test")
        command = Interpolate("echo buildby-%(prop:buildername:~blddef)s")
        rendered = yield self.build.render(command)
        self.assertEqual(rendered, "echo buildby-blddef")

    @defer.inlineCallbacks
    def test_property_colon_plus(self):
        self.props.setProperty("project", "proj1", "test")
        command = Interpolate("echo %(prop:project:+projectdefined)s")
        rendered = yield self.build.render(command)
        self.assertEqual(rendered, "echo projectdefined")

    @defer.inlineCallbacks
    def test_nested_property(self):
        self.props.setProperty("project", "so long!", "test")
        command = Interpolate("echo '%(prop:missing:~%(prop:project)s)s'")
        rendered = yield self.build.render(command)
        self.assertEqual(rendered, "echo 'so long!'")

    @defer.inlineCallbacks
    def test_property_substitute_recursively(self):
        self.props.setProperty("project", "proj1", "test")
        command = Interpolate("echo '%(prop:no_such:-%(prop:project)s)s'")
        rendered = yield self.build.render(command)
        self.assertEqual(rendered, "echo 'proj1'")

    @defer.inlineCallbacks
    def test_property_colon_ternary_present(self):
        self.props.setProperty("project", "proj1", "test")
        command = Interpolate("echo %(prop:project:?:defined:missing)s")
        rendered = yield self.build.render(command)
        self.assertEqual(rendered, "echo defined")

    @defer.inlineCallbacks
    def test_property_colon_ternary_missing(self):
        command = Interpolate("echo %(prop:project:?|defined|missing)s")
        rendered = yield self.build.render(command)
        self.assertEqual(rendered, "echo missing")

    @defer.inlineCallbacks
    def test_property_colon_ternary_hash_true(self):
        self.props.setProperty("project", "winbld", "test")
        command = Interpolate("echo buildby-%(prop:project:#?:T:F)s")
        rendered = yield self.build.render(command)
        self.assertEqual(rendered, "echo buildby-T")

    @defer.inlineCallbacks
    def test_property_colon_ternary_hash_false(self):
        self.props.setProperty("project", "", "test")
        command = Interpolate("echo buildby-%(prop:project:#?|T|F)s")
        rendered = yield self.build.render(command)
        self.assertEqual(rendered, "echo buildby-F")

    @defer.inlineCallbacks
    def test_property_colon_ternary_substitute_recursively_true(self):
        self.props.setProperty("P", "present", "test")
        self.props.setProperty("one", "proj1", "test")
        self.props.setProperty("two", "proj2", "test")
        command = Interpolate("echo '%(prop:P:?|%(prop:one)s|%(prop:two)s)s'")
        rendered = yield self.build.render(command)
        self.assertEqual(rendered, "echo 'proj1'")

    @defer.inlineCallbacks
    def test_property_colon_ternary_substitute_recursively_false(self):
        self.props.setProperty("one", "proj1", "test")
        self.props.setProperty("two", "proj2", "test")
        command = Interpolate("echo '%(prop:P:?|%(prop:one)s|%(prop:two)s)s'")
        rendered = yield self.build.render(command)
        self.assertEqual(rendered, "echo 'proj2'")

    @defer.inlineCallbacks
    def test_property_colon_ternary_substitute_recursively_delimited_true(self):
        self.props.setProperty("P", "present", "test")
        self.props.setProperty("one", "proj1", "test")
        self.props.setProperty("two", "proj2", "test")
        command = Interpolate(
            "echo '%(prop:P:?|%(prop:one:?|true|false)s|%(prop:two:?|false|true)s)s'"
        )
        rendered = yield self.build.render(command)
        self.assertEqual(rendered, "echo 'true'")

    @defer.inlineCallbacks
    def test_property_colon_ternary_substitute_recursively_delimited_false(self):
        self.props.setProperty("one", "proj1", "test")
        self.props.setProperty("two", "proj2", "test")
        command = Interpolate(
            "echo '%(prop:P:?|%(prop:one:?|true|false)s|%(prop:two:?|false|true)s)s'"
        )
        rendered = yield self.build.render(command)
        self.assertEqual(rendered, "echo 'false'")


class TestInterpolateSrc(unittest.TestCase):
    def setUp(self):
        self.props = Properties()
        self.build = FakeBuild(props=self.props)
        sa = FakeSource()
        wfb = FakeSource()
        sc = FakeSource()

        sa.repository = 'cvs://A..'
        sa.codebase = 'cbA'
        sa.project = "Project"
        self.build.sources['cbA'] = sa

        wfb.repository = 'cvs://B..'
        wfb.codebase = 'cbB'
        wfb.project = "Project"
        self.build.sources['cbB'] = wfb

        sc.repository = 'cvs://C..'
        sc.codebase = 'cbC'
        sc.project = None
        self.build.sources['cbC'] = sc

    @defer.inlineCallbacks
    def test_src(self):
        command = Interpolate("echo %(src:cbB:repository)s")
        rendered = yield self.build.render(command)
        self.assertEqual(rendered, "echo cvs://B..")

    @defer.inlineCallbacks
    def test_src_src(self):
        command = Interpolate("echo %(src:cbB:repository)s %(src:cbB:project)s")
        rendered = yield self.build.render(command)
        self.assertEqual(rendered, "echo cvs://B.. Project")

    @defer.inlineCallbacks
    def test_src_attr_empty(self):
        command = Interpolate("echo %(src:cbC:project)s")
        rendered = yield self.build.render(command)
        self.assertEqual(rendered, "echo ")

    @defer.inlineCallbacks
    def test_src_attr_codebase_notfound(self):
        command = Interpolate("echo %(src:unknown_codebase:project)s")
        rendered = yield self.build.render(command)
        self.assertEqual(rendered, "echo ")

    @defer.inlineCallbacks
    def test_src_colon_plus_false(self):
        command = Interpolate("echo '%(src:cbD:project:+defaultrepo)s'")
        rendered = yield self.build.render(command)
        self.assertEqual(rendered, "echo ''")

    @defer.inlineCallbacks
    def test_src_colon_plus_true(self):
        command = Interpolate("echo '%(src:cbB:project:+defaultrepo)s'")
        rendered = yield self.build.render(command)
        self.assertEqual(rendered, "echo 'defaultrepo'")

    @defer.inlineCallbacks
    def test_src_colon_minus(self):
        command = Interpolate("echo %(src:cbB:nonattr:-defaultrepo)s")
        rendered = yield self.build.render(command)
        self.assertEqual(rendered, "echo defaultrepo")

    @defer.inlineCallbacks
    def test_src_colon_minus_false(self):
        command = Interpolate("echo '%(src:cbC:project:-noproject)s'")
        rendered = yield self.build.render(command)
        self.assertEqual(rendered, "echo ''")

    @defer.inlineCallbacks
    def test_src_colon_minus_true(self):
        command = Interpolate("echo '%(src:cbB:project:-noproject)s'")
        rendered = yield self.build.render(command)
        self.assertEqual(rendered, "echo 'Project'")

    @defer.inlineCallbacks
    def test_src_colon_minus_codebase_notfound(self):
        command = Interpolate("echo '%(src:unknown_codebase:project:-noproject)s'")
        rendered = yield self.build.render(command)
        self.assertEqual(rendered, "echo 'noproject'")

    @defer.inlineCallbacks
    def test_src_colon_tilde_true(self):
        command = Interpolate("echo '%(src:cbB:project:~noproject)s'")
        rendered = yield self.build.render(command)
        self.assertEqual(rendered, "echo 'Project'")

    @defer.inlineCallbacks
    def test_src_colon_tilde_false(self):
        command = Interpolate("echo '%(src:cbC:project:~noproject)s'")
        rendered = yield self.build.render(command)
        self.assertEqual(rendered, "echo 'noproject'")

    @defer.inlineCallbacks
    def test_src_colon_tilde_false_src_as_replacement(self):
        command = Interpolate("echo '%(src:cbC:project:~%(src:cbA:project)s)s'")
        rendered = yield self.build.render(command)
        self.assertEqual(rendered, "echo 'Project'")

    @defer.inlineCallbacks
    def test_src_colon_tilde_codebase_notfound(self):
        command = Interpolate("echo '%(src:unknown_codebase:project:~noproject)s'")
        rendered = yield self.build.render(command)
        self.assertEqual(rendered, "echo 'noproject'")


class TestInterpolateKwargs(unittest.TestCase):
    def setUp(self):
        self.props = Properties()
        self.build = FakeBuild(props=self.props)
        sa = FakeSource()

        sa.repository = 'cvs://A..'
        sa.codebase = 'cbA'
        sa.project = None
        sa.branch = "default"
        self.build.sources['cbA'] = sa

    @defer.inlineCallbacks
    def test_kwarg(self):
        command = Interpolate("echo %(kw:repository)s", repository="cvs://A..")
        rendered = yield self.build.render(command)
        self.assertEqual(rendered, "echo cvs://A..")

    @defer.inlineCallbacks
    def test_kwarg_kwarg(self):
        command = Interpolate(
            "echo %(kw:repository)s %(kw:branch)s", repository="cvs://A..", branch="default"
        )
        rendered = yield self.build.render(command)
        self.assertEqual(rendered, "echo cvs://A.. default")

    @defer.inlineCallbacks
    def test_kwarg_not_mapped(self):
        command = Interpolate("echo %(kw:repository)s", project="projectA")
        rendered = yield self.build.render(command)
        self.assertEqual(rendered, "echo ")

    @defer.inlineCallbacks
    def test_kwarg_colon_minus_not_available(self):
        command = Interpolate("echo %(kw:repository)s", project="projectA")
        rendered = yield self.build.render(command)
        self.assertEqual(rendered, "echo ")

    @defer.inlineCallbacks
    def test_kwarg_colon_minus_not_available_default(self):
        command = Interpolate("echo %(kw:repository:-cvs://A..)s", project="projectA")
        rendered = yield self.build.render(command)
        self.assertEqual(rendered, "echo cvs://A..")

    @defer.inlineCallbacks
    def test_kwarg_colon_minus_available(self):
        command = Interpolate("echo %(kw:repository:-cvs://A..)s", repository="cvs://B..")
        rendered = yield self.build.render(command)
        self.assertEqual(rendered, "echo cvs://B..")

    @defer.inlineCallbacks
    def test_kwarg_colon_tilde_true(self):
        command = Interpolate("echo %(kw:repository:~cvs://B..)s", repository="cvs://A..")
        rendered = yield self.build.render(command)
        self.assertEqual(rendered, "echo cvs://A..")

    @defer.inlineCallbacks
    def test_kwarg_colon_tilde_false(self):
        command = Interpolate("echo %(kw:repository:~cvs://B..)s", repository="")
        rendered = yield self.build.render(command)
        self.assertEqual(rendered, "echo cvs://B..")

    @defer.inlineCallbacks
    def test_kwarg_colon_tilde_none(self):
        command = Interpolate("echo %(kw:repository:~cvs://B..)s", repository=None)
        rendered = yield self.build.render(command)
        self.assertEqual(rendered, "echo cvs://B..")

    @defer.inlineCallbacks
    def test_kwarg_colon_plus_false(self):
        command = Interpolate("echo %(kw:repository:+cvs://B..)s", project="project")
        rendered = yield self.build.render(command)
        self.assertEqual(rendered, "echo ")

    @defer.inlineCallbacks
    def test_kwarg_colon_plus_true(self):
        command = Interpolate("echo %(kw:repository:+cvs://B..)s", repository=None)
        rendered = yield self.build.render(command)
        self.assertEqual(rendered, "echo cvs://B..")

    @defer.inlineCallbacks
    def test_kwargs_colon_minus_false_src_as_replacement(self):
        command = Interpolate("echo '%(kw:text:-%(src:cbA:branch)s)s'", notext='ddd')
        rendered = yield self.build.render(command)
        self.assertEqual(rendered, "echo 'default'")

    @defer.inlineCallbacks
    def test_kwargs_renderable(self):
        command = Interpolate("echo '%(kw:test)s'", test=ConstantRenderable('testing'))
        rendered = yield self.build.render(command)
        self.assertEqual(rendered, "echo 'testing'")

    def test_kwargs_deferred(self):
        renderable = DeferredRenderable()
        command = Interpolate("echo '%(kw:test)s'", test=renderable)
        d = self.build.render(command)
        d.addCallback(self.assertEqual, "echo 'testing'")
        renderable.callback('testing')

    def test_kwarg_deferred(self):
        renderable = DeferredRenderable()
        command = Interpolate("echo '%(kw:project)s'", project=renderable)
        d = self.build.render(command)
        d.addCallback(self.assertEqual, "echo 'testing'")
        renderable.callback('testing')

    def test_nested_kwarg_deferred(self):
        renderable = DeferredRenderable()
        command = Interpolate(
            "echo '%(kw:missing:~%(kw:fishy)s)s'", missing=renderable, fishy="so long!"
        )
        d = self.build.render(command)
        d.addCallback(self.assertEqual, "echo 'so long!'")
        renderable.callback(False)
        return d


class TestWithProperties(unittest.TestCase):
    def setUp(self):
        self.props = Properties()
        self.build = FakeBuild(props=self.props)

    def testInvalidParams(self):
        with self.assertRaises(ValueError):
            WithProperties("%s %(foo)s", 1, foo=2)

    @defer.inlineCallbacks
    def testBasic(self):
        # test basic substitution with WithProperties
        self.props.setProperty("revision", "47", "test")
        command = WithProperties("build-%s.tar.gz", "revision")
        res = yield self.build.render(command)
        self.assertEqual(res, "build-47.tar.gz")

    @defer.inlineCallbacks
    def testDict(self):
        # test dict-style substitution with WithProperties
        self.props.setProperty("other", "foo", "test")
        command = WithProperties("build-%(other)s.tar.gz")
        res = yield self.build.render(command)
        self.assertEqual(res, "build-foo.tar.gz")

    @defer.inlineCallbacks
    def testDictColonMinus(self):
        # test dict-style substitution with WithProperties
        self.props.setProperty("prop1", "foo", "test")
        command = WithProperties("build-%(prop1:-empty)s-%(prop2:-empty)s.tar.gz")
        res = yield self.build.render(command)
        self.assertEqual(res, "build-foo-empty.tar.gz")

    @defer.inlineCallbacks
    def testDictColonPlus(self):
        # test dict-style substitution with WithProperties
        self.props.setProperty("prop1", "foo", "test")
        command = WithProperties("build-%(prop1:+exists)s-%(prop2:+exists)s.tar.gz")
        res = yield self.build.render(command)
        self.assertEqual(res, "build-exists-.tar.gz")

    @defer.inlineCallbacks
    def testEmpty(self):
        # None should render as ''
        self.props.setProperty("empty", None, "test")
        command = WithProperties("build-%(empty)s.tar.gz")
        res = yield self.build.render(command)
        self.assertEqual(res, "build-.tar.gz")

    @defer.inlineCallbacks
    def testRecursiveList(self):
        self.props.setProperty("x", 10, "test")
        self.props.setProperty("y", 20, "test")
        command = [WithProperties("%(x)s %(y)s"), "and", WithProperties("%(y)s %(x)s")]
        res = yield self.build.render(command)
        self.assertEqual(res, ["10 20", "and", "20 10"])

    @defer.inlineCallbacks
    def testRecursiveTuple(self):
        self.props.setProperty("x", 10, "test")
        self.props.setProperty("y", 20, "test")
        command = (WithProperties("%(x)s %(y)s"), "and", WithProperties("%(y)s %(x)s"))
        res = yield self.build.render(command)
        self.assertEqual(res, ("10 20", "and", "20 10"))

    @defer.inlineCallbacks
    def testRecursiveDict(self):
        self.props.setProperty("x", 10, "test")
        self.props.setProperty("y", 20, "test")
        command = {WithProperties("%(x)s %(y)s"): WithProperties("%(y)s %(x)s")}
        res = yield self.build.render(command)
        self.assertEqual(res, {"10 20": "20 10"})

    @defer.inlineCallbacks
    def testLambdaSubst(self):
        command = WithProperties('%(foo)s', foo=lambda _: 'bar')
        res = yield self.build.render(command)
        self.assertEqual(res, 'bar')

    @defer.inlineCallbacks
    def testLambdaHasattr(self):
        command = WithProperties('%(foo)s', foo=lambda b: (b.hasProperty('x') and 'x') or 'y')
        res = yield self.build.render(command)
        self.assertEqual(res, 'y')

    @defer.inlineCallbacks
    def testLambdaOverride(self):
        self.props.setProperty('x', 10, 'test')
        command = WithProperties('%(x)s', x=lambda _: 20)
        res = yield self.build.render(command)
        self.assertEqual(res, '20')

    def testLambdaCallable(self):
        with self.assertRaises(ValueError):
            WithProperties('%(foo)s', foo='bar')

    @defer.inlineCallbacks
    def testLambdaUseExisting(self):
        self.props.setProperty('x', 10, 'test')
        self.props.setProperty('y', 20, 'test')
        command = WithProperties(
            '%(z)s', z=lambda props: props.getProperty('x') + props.getProperty('y')
        )
        res = yield self.build.render(command)
        self.assertEqual(res, '30')

    @defer.inlineCallbacks
    def testColon(self):
        self.props.setProperty('some:property', 10, 'test')
        command = WithProperties('%(some:property:-with-default)s')
        res = yield self.build.render(command)
        self.assertEqual(res, '10')

    @defer.inlineCallbacks
    def testColon_default(self):
        command = WithProperties('%(some:property:-with-default)s')
        res = yield self.build.render(command)
        self.assertEqual(res, 'with-default')

    @defer.inlineCallbacks
    def testColon_colon(self):
        command = WithProperties('%(some:property:-with:default)s')
        res = yield self.build.render(command)
        self.assertEqual(res, 'with:default')


class TestProperties(unittest.TestCase):
    def setUp(self):
        self.props = Properties()

    def testDictBehavior(self):
        # note that dictionary-like behavior is deprecated and not exposed to
        # users!
        self.props.setProperty("do-tests", 1, "scheduler")
        self.props.setProperty("do-install", 2, "scheduler")

        self.assertTrue('do-tests' in self.props)
        self.assertEqual(self.props['do-tests'], 1)
        self.assertEqual(self.props['do-install'], 2)
        with self.assertRaises(KeyError):
            self.props['do-nothing']
        self.assertEqual(self.props.getProperty('do-install'), 2)
        self.assertIn('do-tests', self.props)
        self.assertNotIn('missing-do-tests', self.props)

    def testAsList(self):
        self.props.setProperty("happiness", 7, "builder")
        self.props.setProperty("flames", True, "tester")

        self.assertEqual(
            sorted(self.props.asList()), [('flames', True, 'tester'), ('happiness', 7, 'builder')]
        )

    def testAsDict(self):
        self.props.setProperty("msi_filename", "product.msi", 'packager')
        self.props.setProperty("dmg_filename", "product.dmg", 'packager')

        self.assertEqual(
            self.props.asDict(),
            {
                "msi_filename": ('product.msi', 'packager'),
                "dmg_filename": ('product.dmg', 'packager'),
            },
        )

    def testUpdate(self):
        self.props.setProperty("x", 24, "old")
        newprops = {'a': 1, 'b': 2}
        self.props.update(newprops, "new")

        self.assertEqual(self.props.getProperty('x'), 24)
        self.assertEqual(self.props.getPropertySource('x'), 'old')
        self.assertEqual(self.props.getProperty('a'), 1)
        self.assertEqual(self.props.getPropertySource('a'), 'new')

    def testUpdateRuntime(self):
        self.props.setProperty("x", 24, "old")
        newprops = {'a': 1, 'b': 2}
        self.props.update(newprops, "new", runtime=True)

        self.assertEqual(self.props.getProperty('x'), 24)
        self.assertEqual(self.props.getPropertySource('x'), 'old')
        self.assertEqual(self.props.getProperty('a'), 1)
        self.assertEqual(self.props.getPropertySource('a'), 'new')
        self.assertEqual(self.props.runtime, set(['a', 'b']))

    def testUpdateFromProperties(self):
        self.props.setProperty("a", 94, "old")
        self.props.setProperty("x", 24, "old")
        newprops = Properties()
        newprops.setProperty('a', 1, "new")
        newprops.setProperty('b', 2, "new")
        self.props.updateFromProperties(newprops)

        self.assertEqual(self.props.getProperty('x'), 24)
        self.assertEqual(self.props.getPropertySource('x'), 'old')
        self.assertEqual(self.props.getProperty('a'), 1)
        self.assertEqual(self.props.getPropertySource('a'), 'new')

    def testUpdateFromPropertiesNoRuntime(self):
        self.props.setProperty("a", 94, "old")
        self.props.setProperty("b", 84, "old")
        self.props.setProperty("x", 24, "old")
        newprops = Properties()
        newprops.setProperty('a', 1, "new", runtime=True)
        newprops.setProperty('b', 2, "new", runtime=False)
        newprops.setProperty('c', 3, "new", runtime=True)
        newprops.setProperty('d', 3, "new", runtime=False)
        self.props.updateFromPropertiesNoRuntime(newprops)

        self.assertEqual(self.props.getProperty('a'), 94)
        self.assertEqual(self.props.getPropertySource('a'), 'old')
        self.assertEqual(self.props.getProperty('b'), 2)
        self.assertEqual(self.props.getPropertySource('b'), 'new')
        self.assertEqual(self.props.getProperty('c'), None)  # not updated
        self.assertEqual(self.props.getProperty('d'), 3)
        self.assertEqual(self.props.getPropertySource('d'), 'new')
        self.assertEqual(self.props.getProperty('x'), 24)
        self.assertEqual(self.props.getPropertySource('x'), 'old')

    def test_setProperty_notJsonable(self):
        with self.assertRaises(TypeError):
            self.props.setProperty("project", object, "test")

    # IProperties methods

    def test_getProperty(self):
        self.props.properties['p1'] = (['p', 1], 'test')
        self.assertEqual(self.props.getProperty('p1'), ['p', 1])

    def test_getProperty_default_None(self):
        self.assertEqual(self.props.getProperty('p1'), None)

    def test_getProperty_default(self):
        self.assertEqual(self.props.getProperty('p1', 2), 2)

    def test_hasProperty_false(self):
        self.assertFalse(self.props.hasProperty('x'))

    def test_hasProperty_true(self):
        self.props.properties['x'] = (False, 'test')
        self.assertTrue(self.props.hasProperty('x'))

    def test_has_key_false(self):
        self.assertFalse('x' in self.props)

    def test_setProperty(self):
        self.props.setProperty('x', 'y', 'test')
        self.assertEqual(self.props.properties['x'], ('y', 'test'))
        self.assertNotIn('x', self.props.runtime)

    def test_setProperty_runtime(self):
        self.props.setProperty('x', 'y', 'test', runtime=True)
        self.assertEqual(self.props.properties['x'], ('y', 'test'))
        self.assertIn('x', self.props.runtime)

    def test_setProperty_no_source(self):
        # pylint: disable=no-value-for-parameter
        with self.assertRaises(TypeError):
            self.props.setProperty('x', 'y')

    def test_getProperties(self):
        self.assertIdentical(self.props.getProperties(), self.props)

    def test_getBuild(self):
        self.assertIdentical(self.props.getBuild(), self.props.build)

    def test_unset_sourcestamps(self):
        with self.assertRaises(AttributeError):
            self.props.sourcestamps()

    def test_unset_changes(self):
        with self.assertRaises(AttributeError):
            self.props.changes()
        with self.assertRaises(AttributeError):
            self.props.files()

    def test_build_attributes(self):
        build = FakeBuild(self.props)
        change = TempChange({'author': 'me', 'files': ['main.c']})
        ss = TempSourceStamp({'branch': 'master'})
        ss.changes = [change]
        build.sources[''] = ss
        self.assertEqual(self.props.sourcestamps[0]['branch'], 'master')
        self.assertEqual(self.props.changes[0]['author'], 'me')
        self.assertEqual(self.props.files[0], 'main.c')

    def test_own_attributes(self):
        self.props.sourcestamps = [{'branch': 'master'}]
        self.props.changes = [{'author': 'me', 'files': ['main.c']}]
        self.assertEqual(self.props.sourcestamps[0]['branch'], 'master')
        self.assertEqual(self.props.changes[0]['author'], 'me')
        self.assertEqual(self.props.files[0], 'main.c')

    @defer.inlineCallbacks
    def test_render(self):
        @implementer(IRenderable)
        class Renderable:
            def getRenderingFor(self, props):
                return props.getProperty('x') + 'z'

        self.props.setProperty('x', 'y', 'test')
        res = yield self.props.render(Renderable())
        self.assertEqual(res, 'yz')


class MyPropertiesThing(PropertiesMixin):
    set_runtime_properties = True

    def getProperties(self):
        return self.properties


class TestPropertiesMixin(unittest.TestCase):
    def setUp(self):
        self.mp = MyPropertiesThing()
        self.mp.properties = mock.Mock()

    def test_getProperty(self):
        self.mp.getProperty('abc')
        self.mp.properties.getProperty.assert_called_with('abc', None)

    def xtest_getProperty_default(self):
        self.mp.getProperty('abc', 'def')
        self.mp.properties.getProperty.assert_called_with('abc', 'def')

    def test_hasProperty(self):
        self.mp.properties.hasProperty.return_value = True
        self.assertTrue(self.mp.hasProperty('abc'))
        self.mp.properties.hasProperty.assert_called_with('abc')

    def test_has_key(self):
        self.mp.properties.hasProperty.return_value = True
        # getattr because pep8 doesn't like calls to has_key
        self.assertTrue(self.mp.has_key('abc'))
        self.mp.properties.hasProperty.assert_called_with('abc')

    def test_setProperty(self):
        self.mp.setProperty('abc', 'def', 'src')
        self.mp.properties.setProperty.assert_called_with('abc', 'def', 'src', runtime=True)

    def test_setProperty_no_source(self):
        # this compatibility is maintained for old code
        self.mp.setProperty('abc', 'def')
        self.mp.properties.setProperty.assert_called_with('abc', 'def', 'Unknown', runtime=True)

    def test_render(self):
        self.mp.render([1, 2])
        self.mp.properties.render.assert_called_with([1, 2])


class TestProperty(unittest.TestCase):
    def setUp(self):
        self.props = Properties()
        self.build = FakeBuild(props=self.props)

    @defer.inlineCallbacks
    def testIntProperty(self):
        self.props.setProperty("do-tests", 1, "scheduler")
        value = Property("do-tests")

        res = yield self.build.render(value)
        self.assertEqual(res, 1)

    @defer.inlineCallbacks
    def testStringProperty(self):
        self.props.setProperty("do-tests", "string", "scheduler")
        value = Property("do-tests")

        res = yield self.build.render(value)
        self.assertEqual(res, "string")

    @defer.inlineCallbacks
    def testMissingProperty(self):
        value = Property("do-tests")

        res = yield self.build.render(value)
        self.assertEqual(res, None)

    @defer.inlineCallbacks
    def testDefaultValue(self):
        value = Property("do-tests", default="Hello!")

        res = yield self.build.render(value)
        self.assertEqual(res, "Hello!")

    @defer.inlineCallbacks
    def testDefaultValueNested(self):
        self.props.setProperty("xxx", 'yyy', "scheduler")
        value = Property("do-tests", default=WithProperties("a-%(xxx)s-b"))

        res = yield self.build.render(value)
        self.assertEqual(res, "a-yyy-b")

    @defer.inlineCallbacks
    def testIgnoreDefaultValue(self):
        self.props.setProperty("do-tests", "string", "scheduler")
        value = Property("do-tests", default="Hello!")

        res = yield self.build.render(value)
        self.assertEqual(res, "string")

    @defer.inlineCallbacks
    def testIgnoreFalseValue(self):
        self.props.setProperty("do-tests-string", "", "scheduler")
        self.props.setProperty("do-tests-int", 0, "scheduler")
        self.props.setProperty("do-tests-list", [], "scheduler")
        self.props.setProperty("do-tests-None", None, "scheduler")

        value = [
            Property("do-tests-string", default="Hello!"),
            Property("do-tests-int", default="Hello!"),
            Property("do-tests-list", default="Hello!"),
            Property("do-tests-None", default="Hello!"),
        ]

        res = yield self.build.render(value)
        self.assertEqual(res, ["Hello!"] * 4)

    @defer.inlineCallbacks
    def testDefaultWhenFalse(self):
        self.props.setProperty("do-tests-string", "", "scheduler")
        self.props.setProperty("do-tests-int", 0, "scheduler")
        self.props.setProperty("do-tests-list", [], "scheduler")
        self.props.setProperty("do-tests-None", None, "scheduler")

        value = [
            Property("do-tests-string", default="Hello!", defaultWhenFalse=False),
            Property("do-tests-int", default="Hello!", defaultWhenFalse=False),
            Property("do-tests-list", default="Hello!", defaultWhenFalse=False),
            Property("do-tests-None", default="Hello!", defaultWhenFalse=False),
        ]

        res = yield self.build.render(value)
        self.assertEqual(res, ["", 0, [], None])

    def testDeferredDefault(self):
        default = DeferredRenderable()
        value = Property("no-such-property", default)
        d = self.build.render(value)
        d.addCallback(self.assertEqual, "default-value")
        default.callback("default-value")
        return d

    @defer.inlineCallbacks
    def testFlattenList(self):
        self.props.setProperty("do-tests", "string", "scheduler")
        value = FlattenList([Property("do-tests"), ["bla"]])

        res = yield self.build.render(value)
        self.assertEqual(res, ["string", "bla"])

    @defer.inlineCallbacks
    def testFlattenListAdd(self):
        self.props.setProperty("do-tests", "string", "scheduler")
        value = FlattenList([Property("do-tests"), ["bla"]])
        value = value + FlattenList([Property("do-tests"), ["bla"]])

        res = yield self.build.render(value)
        self.assertEqual(res, ["string", "bla", "string", "bla"])

    @defer.inlineCallbacks
    def testFlattenListAdd2(self):
        self.props.setProperty("do-tests", "string", "scheduler")
        value = FlattenList([Property("do-tests"), ["bla"]])
        value = value + [Property('do-tests'), ['bla']]  # noqa: RUF005

        res = yield self.build.render(value)
        self.assertEqual(res, ["string", "bla", "string", "bla"])

    @defer.inlineCallbacks
    def testCompEq(self):
        self.props.setProperty("do-tests", "string", "scheduler")
        result = yield self.build.render(Property("do-tests") == "string")
        self.assertEqual(result, True)

    @defer.inlineCallbacks
    def testCompNe(self):
        self.props.setProperty("do-tests", "not-string", "scheduler")
        result = yield self.build.render(Property("do-tests") != "string")
        self.assertEqual(result, True)

    @defer.inlineCallbacks
    def testCompLt(self):
        self.props.setProperty("do-tests", 1, "scheduler")
        x = Property("do-tests") < 2
        self.assertEqual(repr(x), 'Property(do-tests) < 2')
        result = yield self.build.render(x)
        self.assertEqual(result, True)

    @defer.inlineCallbacks
    def testCompLe(self):
        self.props.setProperty("do-tests", 1, "scheduler")
        result = yield self.build.render(Property("do-tests") <= 2)
        self.assertEqual(result, True)

    @defer.inlineCallbacks
    def testCompGt(self):
        self.props.setProperty("do-tests", 3, "scheduler")
        result = yield self.build.render(Property("do-tests") > 2)
        self.assertEqual(result, True)

    @defer.inlineCallbacks
    def testCompGe(self):
        self.props.setProperty("do-tests", 3, "scheduler")
        result = yield self.build.render(Property("do-tests") >= 2)
        self.assertEqual(result, True)

    @defer.inlineCallbacks
    def testStringCompEq(self):
        self.props.setProperty("do-tests", "string", "scheduler")
        test_string = "string"
        result = yield self.build.render(test_string == Property("do-tests"))
        self.assertEqual(result, True)

    @defer.inlineCallbacks
    def testIntCompLe(self):
        self.props.setProperty("do-tests", 1, "scheduler")
        test_int = 1
        result = yield self.build.render(test_int <= Property("do-tests"))
        self.assertEqual(result, True)

    @defer.inlineCallbacks
    def testPropCompGe(self):
        self.props.setProperty("do-tests", 1, "scheduler")
        result = yield self.build.render(Property("do-tests") >= Property("do-tests"))
        self.assertEqual(result, True)

    @defer.inlineCallbacks
    def testPropAdd(self):
        self.props.setProperty("do-tests", 1, "scheduler")
        result = yield self.build.render(Property("do-tests") + Property("do-tests"))
        self.assertEqual(result, 2)

    @defer.inlineCallbacks
    def testPropSub(self):
        self.props.setProperty("do-tests", 1, "scheduler")
        result = yield self.build.render(Property("do-tests") - Property("do-tests"))
        self.assertEqual(result, 0)

    @defer.inlineCallbacks
    def testPropDiv(self):
        self.props.setProperty("do-tests", 1, "scheduler")
        self.props.setProperty("do-tests2", 3, "scheduler")
        result = yield self.build.render(Property("do-tests") / Property("do-tests2"))
        self.assertEqual(result, 1 / 3)

    @defer.inlineCallbacks
    def testPropFDiv(self):
        self.props.setProperty("do-tests", 5, "scheduler")
        self.props.setProperty("do-tests2", 2, "scheduler")
        result = yield self.build.render(Property("do-tests") // Property("do-tests2"))
        self.assertEqual(result, 2)

    @defer.inlineCallbacks
    def testPropMod(self):
        self.props.setProperty("do-tests", 5, "scheduler")
        self.props.setProperty("do-tests2", 3, "scheduler")
        result = yield self.build.render(Property("do-tests") % Property("do-tests2"))
        self.assertEqual(result, 2)

    @defer.inlineCallbacks
    def testPropMult(self):
        self.props.setProperty("do-tests", 2, "scheduler")
        result = yield self.build.render(Property("do-tests") * Interpolate("%(prop:do-tests)s"))
        self.assertEqual(result, '22')

    @defer.inlineCallbacks
    def testPropIn(self):
        self.props.setProperty("do-tests", 2, "scheduler")
        result = yield self.build.render(Property("do-tests").in_([1, 2]))
        self.assertEqual(result, True)

    @defer.inlineCallbacks
    def testPropIn2(self):
        self.props.setProperty("do-tests", 2, "scheduler")
        result = yield self.build.render(Property("do-tests").in_([1, 3]))
        self.assertEqual(result, False)


class TestRenderableAdapters(unittest.TestCase):
    """
    Tests for list, tuple and dict renderers.
    """

    def setUp(self):
        self.props = Properties()
        self.build = FakeBuild(props=self.props)

    def test_list_deferred(self):
        r1 = DeferredRenderable()
        r2 = DeferredRenderable()
        d = self.build.render([r1, r2])
        d.addCallback(self.assertEqual, ["lispy", "lists"])
        r2.callback("lists")
        r1.callback("lispy")
        return d

    def test_tuple_deferred(self):
        r1 = DeferredRenderable()
        r2 = DeferredRenderable()
        d = self.build.render((r1, r2))
        d.addCallback(self.assertEqual, ("totally", "tupled"))
        r2.callback("tupled")
        r1.callback("totally")
        return d

    def test_dict(self):
        r1 = DeferredRenderable()
        r2 = DeferredRenderable()
        k1 = DeferredRenderable()
        k2 = DeferredRenderable()
        d = self.build.render({k1: r1, k2: r2})
        d.addCallback(self.assertEqual, {"lock": "load", "dict": "lookup"})
        k1.callback("lock")
        r1.callback("load")
        k2.callback("dict")
        r2.callback("lookup")
        return d


class Renderer(unittest.TestCase):
    def setUp(self):
        self.props = Properties()
        self.build = FakeBuild(props=self.props)

    @defer.inlineCallbacks
    def test_renderer(self):
        self.props.setProperty("x", "X", "test")

        def rend(p):
            return f"x{p.getProperty('x')}x"

        res = yield self.build.render(renderer(rend))
        self.assertEqual('xXx', res)

    @defer.inlineCallbacks
    def test_renderer_called(self):
        # it's tempting to try to call the decorated function.  Don't do that.
        # It's not a function anymore.

        def rend(p):
            return 'x'

        with self.assertRaises(TypeError):
            yield self.build.render(renderer(rend)('y'))

    @defer.inlineCallbacks
    def test_renderer_decorator(self):
        self.props.setProperty("x", "X", "test")

        @renderer
        def rend(p):
            return f"x{p.getProperty('x')}x"

        res = yield self.build.render(rend)
        self.assertEqual('xXx', res)

    @defer.inlineCallbacks
    def test_renderer_deferred(self):
        self.props.setProperty("x", "X", "test")

        def rend(p):
            return defer.succeed(f"y{p.getProperty('x')}y")

        res = yield self.build.render(renderer(rend))
        self.assertEqual('yXy', res)

    @defer.inlineCallbacks
    def test_renderer_fails(self):
        @defer.inlineCallbacks
        def rend(p):
            raise RuntimeError("oops")

        with self.assertRaises(RuntimeError):
            yield self.build.render(renderer(rend))

    @defer.inlineCallbacks
    def test_renderer_recursive(self):
        self.props.setProperty("x", "X", "test")

        def rend(p):
            return Interpolate("x%(prop:x)sx")

        ret = yield self.build.render(renderer(rend))
        self.assertEqual('xXx', ret)

    def test_renderer_repr(self):
        @renderer
        def myrend(p):
            pass

        self.assertIn('renderer(', repr(myrend))
        # py3 and py2 do not have the same way of repr functions
        # but they always contain the name of function
        self.assertIn('myrend', repr(myrend))

    @defer.inlineCallbacks
    def test_renderer_with_state(self):
        self.props.setProperty("x", "X", "test")

        def rend(p, arg, kwarg='y'):
            return f"x-{p.getProperty('x')}-{arg}-{kwarg}"

        res = yield self.build.render(renderer(rend).withArgs('a', kwarg='kw'))
        self.assertEqual('x-X-a-kw', res)

    @defer.inlineCallbacks
    def test_renderer_with_state_called(self):
        # it's tempting to try to call the decorated function.  Don't do that.
        # It's not a function anymore.

        def rend(p, arg, kwarg='y'):
            return 'x'

        with self.assertRaises(TypeError):
            rend_with_args = renderer(rend).withArgs('a', kwarg='kw')
            yield self.build.render(rend_with_args('y'))

    @defer.inlineCallbacks
    def test_renderer_with_state_renders_args(self):
        self.props.setProperty("x", "X", "test")
        self.props.setProperty('arg', 'ARG', 'test2')
        self.props.setProperty('kw', 'KW', 'test3')

        def rend(p, arg, kwarg='y'):
            return f"x-{p.getProperty('x')}-{arg}-{kwarg}"

        res = yield self.build.render(
            renderer(rend).withArgs(Property('arg'), kwarg=Property('kw'))
        )
        self.assertEqual('x-X-ARG-KW', res)

    @defer.inlineCallbacks
    def test_renderer_decorator_with_state(self):
        self.props.setProperty("x", "X", "test")

        @renderer
        def rend(p, arg, kwarg='y'):
            return f"x-{p.getProperty('x')}-{arg}-{kwarg}"

        res = yield self.build.render(rend.withArgs('a', kwarg='kw'))
        self.assertEqual('x-X-a-kw', res)

    @defer.inlineCallbacks
    def test_renderer_decorator_with_state_does_not_share_state(self):
        self.props.setProperty("x", "X", "test")

        @renderer
        def rend(p, *args, **kwargs):
            return f"x-{p.getProperty('x')}-{args!s}-{kwargs!s}"

        rend1 = rend.withArgs('a', kwarg1='kw1')
        rend2 = rend.withArgs('b', kwarg2='kw2')

        res1 = yield self.build.render(rend1)
        res2 = yield self.build.render(rend2)

        self.assertEqual('x-X-(\'a\',)-{\'kwarg1\': \'kw1\'}', res1)
        self.assertEqual('x-X-(\'b\',)-{\'kwarg2\': \'kw2\'}', res2)

    @defer.inlineCallbacks
    def test_renderer_deferred_with_state(self):
        self.props.setProperty("x", "X", "test")

        def rend(p, arg, kwarg='y'):
            return defer.succeed(f"x-{p.getProperty('x')}-{arg}-{kwarg}")

        res = yield self.build.render(renderer(rend).withArgs('a', kwarg='kw'))
        self.assertEqual('x-X-a-kw', res)

    @defer.inlineCallbacks
    def test_renderer_fails_with_state(self):
        self.props.setProperty("x", "X", "test")

        def rend(p, arg, kwarg='y'):
            raise RuntimeError('oops')

        with self.assertRaises(RuntimeError):
            yield self.build.render(renderer(rend).withArgs('a', kwarg='kw'))

    @defer.inlineCallbacks
    def test_renderer_recursive_with_state(self):
        self.props.setProperty("x", "X", "test")

        def rend(p, arg, kwarg='y'):
            return Interpolate('x-%(prop:x)s-%(kw:arg)s-%(kw:kwarg)s', arg=arg, kwarg=kwarg)

        res = yield self.build.render(renderer(rend).withArgs('a', kwarg='kw'))
        self.assertEqual('x-X-a-kw', res)

    def test_renderer_repr_with_state(self):
        @renderer
        def rend(p):
            pass

        rend = rend.withArgs('a', kwarg='kw')  # pylint: disable=assignment-from-no-return

        self.assertIn('renderer(', repr(rend))
        # py3 and py2 do not have the same way of repr functions
        # but they always contain the name of function
        self.assertIn('args=[\'a\']', repr(rend))
        self.assertIn('kwargs={\'kwarg\': \'kw\'}', repr(rend))

    @defer.inlineCallbacks
    def test_interpolate_worker(self):
        self.build.workerforbuilder.worker.info.setProperty('test', 'testvalue', 'Worker')
        rend = yield self.build.render(Interpolate("%(worker:test)s"))
        self.assertEqual(rend, "testvalue")


class Compare(unittest.TestCase):
    def test_WithProperties_lambda(self):
        self.assertNotEqual(
            WithProperties("%(key)s", key=lambda p: 'val'),
            WithProperties("%(key)s", key=lambda p: 'val'),
        )

        def rend(p):
            return "val"

        self.assertEqual(WithProperties("%(key)s", key=rend), WithProperties("%(key)s", key=rend))
        self.assertNotEqual(
            WithProperties("%(key)s", key=rend), WithProperties("%(key)s", otherkey=rend)
        )

    def test_WithProperties_positional(self):
        self.assertNotEqual(WithProperties("%s", 'key'), WithProperties("%s", 'otherkey'))
        self.assertEqual(WithProperties("%s", 'key'), WithProperties("%s", 'key'))
        self.assertNotEqual(WithProperties("%s", 'key'), WithProperties("k%s", 'key'))

    def test_Interpolate_constant(self):
        self.assertNotEqual(Interpolate('some text here'), Interpolate('and other text there'))
        self.assertEqual(Interpolate('some text here'), Interpolate('some text here'))

    def test_Interpolate_positional(self):
        self.assertNotEqual(
            Interpolate('%s %s', "test", "text"), Interpolate('%s %s', "other", "text")
        )
        self.assertEqual(Interpolate('%s %s', "test", "text"), Interpolate('%s %s', "test", "text"))

    def test_Interpolate_kwarg(self):
        self.assertNotEqual(
            Interpolate("%(kw:test)s", test=object(), other=2),
            Interpolate("%(kw:test)s", test=object(), other=2),
        )
        self.assertEqual(
            Interpolate('testing: %(kw:test)s', test="test", other=3),
            Interpolate('testing: %(kw:test)s', test="test", other=3),
        )

    def test_Interpolate_worker(self):
        self.assertEqual(
            Interpolate('testing: %(worker:test)s'), Interpolate('testing: %(worker:test)s')
        )

    def test_renderer(self):
        self.assertNotEqual(renderer(lambda p: 'val'), renderer(lambda p: 'val'))

        def rend(p):
            return "val"

        self.assertEqual(renderer(rend), renderer(rend))

    def test_Lookup_simple(self):
        self.assertNotEqual(
            _Lookup({'test': 5, 'other': 6}, 'other'), _Lookup({'test': 5, 'other': 6}, 'test')
        )
        self.assertEqual(
            _Lookup({'test': 5, 'other': 6}, 'test'), _Lookup({'test': 5, 'other': 6}, 'test')
        )

    def test_Lookup_default(self):
        self.assertNotEqual(
            _Lookup({'test': 5, 'other': 6}, 'test', default='default'),
            _Lookup({'test': 5, 'other': 6}, 'test'),
        )
        self.assertEqual(
            _Lookup({'test': 5, 'other': 6}, 'test', default='default'),
            _Lookup({'test': 5, 'other': 6}, 'test', default='default'),
        )

    def test_Lookup_defaultWhenFalse(self):
        self.assertNotEqual(
            _Lookup({'test': 5, 'other': 6}, 'test', defaultWhenFalse=False),
            _Lookup({'test': 5, 'other': 6}, 'test'),
        )
        self.assertNotEqual(
            _Lookup({'test': 5, 'other': 6}, 'test', defaultWhenFalse=False),
            _Lookup({'test': 5, 'other': 6}, 'test', defaultWhenFalse=True),
        )
        self.assertEqual(
            _Lookup({'test': 5, 'other': 6}, 'test', defaultWhenFalse=True),
            _Lookup({'test': 5, 'other': 6}, 'test', defaultWhenFalse=True),
        )
        self.assertEqual(
            _Lookup({'test': 5, 'other': 6}, 'test'),
            _Lookup({'test': 5, 'other': 6}, 'test', defaultWhenFalse=True),
        )

    def test_Lookup_hasKey(self):
        self.assertNotEqual(
            _Lookup({'test': 5, 'other': 6}, 'test', hasKey=None),
            _Lookup({'test': 5, 'other': 6}, 'test'),
        )
        self.assertNotEqual(
            _Lookup({'test': 5, 'other': 6}, 'test', hasKey='has-key'),
            _Lookup({'test': 5, 'other': 6}, 'test'),
        )
        self.assertNotEqual(
            _Lookup({'test': 5, 'other': 6}, 'test', hasKey='has-key'),
            _Lookup({'test': 5, 'other': 6}, 'test', hasKey='other-key'),
        )
        self.assertEqual(
            _Lookup({'test': 5, 'other': 6}, 'test', hasKey='has-key'),
            _Lookup({'test': 5, 'other': 6}, 'test', hasKey='has-key'),
        )

    def test_Lookup_elideNoneAs(self):
        self.assertEqual(
            _Lookup({'test': 5, 'other': 6}, 'test', elideNoneAs=None),
            _Lookup({'test': 5, 'other': 6}, 'test'),
        )
        self.assertNotEqual(
            _Lookup({'test': 5, 'other': 6}, 'test', elideNoneAs=''),
            _Lookup({'test': 5, 'other': 6}, 'test'),
        )
        self.assertNotEqual(
            _Lookup({'test': 5, 'other': 6}, 'test', elideNoneAs='got None'),
            _Lookup({'test': 5, 'other': 6}, 'test', elideNoneAs=''),
        )
        self.assertEqual(
            _Lookup({'test': 5, 'other': 6}, 'test', elideNoneAs='got None'),
            _Lookup({'test': 5, 'other': 6}, 'test', elideNoneAs='got None'),
        )

    def test_Lazy(self):
        self.assertNotEqual(_Lazy(5), _Lazy(6))
        self.assertEqual(_Lazy(5), _Lazy(5))

    def test_SourceStampDict(self):
        self.assertNotEqual(_SourceStampDict('binary'), _SourceStampDict('library'))
        self.assertEqual(_SourceStampDict('binary'), _SourceStampDict('binary'))


class TestTransform(unittest.TestCase, ConfigErrorsMixin):
    def setUp(self):
        self.props = Properties(propname='propvalue')

    def test_invalid_first_arg(self):
        with self.assertRaisesConfigError(
            "function given to Transform neither callable nor renderable"
        ):
            Transform(None)

    @defer.inlineCallbacks
    def test_argless(self):
        t = Transform(lambda: 'abc')
        res = yield self.props.render(t)
        self.assertEqual(res, 'abc')

    @defer.inlineCallbacks
    def test_argless_renderable(self):
        @renderer
        def function(iprops):
            return lambda: iprops.getProperty('propname')

        t = Transform(function)
        res = yield self.props.render(t)
        self.assertEqual(res, 'propvalue')

    @defer.inlineCallbacks
    def test_args(self):
        t = Transform(lambda x, y: x + '|' + y, 'abc', Property('propname'))
        res = yield self.props.render(t)
        self.assertEqual(res, 'abc|propvalue')

    @defer.inlineCallbacks
    def test_kwargs(self):
        t = Transform(lambda x, y: x + '|' + y, x='abc', y=Property('propname'))
        res = yield self.props.render(t)
        self.assertEqual(res, 'abc|propvalue')

    def test_deferred(self):
        function = DeferredRenderable()
        arg = DeferredRenderable()
        kwarg = DeferredRenderable()

        t = Transform(function, arg, y=kwarg)
        d = self.props.render(t)
        d.addCallback(self.assertEqual, 'abc|def')

        function.callback(lambda x, y: x + '|' + y)
        arg.callback('abc')
        kwarg.callback('def')

        return d



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

from typing import Any
from typing import cast

from parameterized import parameterized
from twisted.trial import unittest

from buildbot.data import codebase_branches
from buildbot.data import resultspec
from buildbot.master import BuildMaster
from buildbot.test import fakedb
from buildbot.test.fake import fakemaster
from buildbot.test.reactor import TestReactorMixin
from buildbot.test.util import endpoint
from buildbot.test.util import interfaces
from buildbot.util.twisted import async_to_deferred


class CodebaseBranchEndpoint(endpoint.EndpointMixin, unittest.TestCase):
    endpointClass = codebase_branches.CodebaseBranchEndpoint
    resourceTypeClass = codebase_branches.CodebaseBranch

    @async_to_deferred
    async def setUp(self) -> None:  # type: ignore[override]
        await self.setUpEndpoint()
        await self.master.db.insert_test_data([
            fakedb.Project(id=7, name='fake_project7'),
            fakedb.Codebase(id=13, projectid=7, name='codebase1'),
            fakedb.CodebaseCommit(id=110, codebaseid=13),
            fakedb.CodebaseBranch(id=220, codebaseid=13, name='branch1', commitid=110),
        ])

    @async_to_deferred
    async def test_get_existing(self) -> None:
        branch = await self.callGet(('branches', 220))

        self.validateData(branch)
        self.assertEqual(
            branch,
            {
                'branchid': 220,
                'codebaseid': 13,
                'commitid': 110,
                'last_timestamp': 1234567,
                'name': 'branch1',
            },
        )

    @async_to_deferred
    async def test_get_missing(self) -> None:
        branch = await self.callGet(('branches', 234))
        self.assertIsNone(branch)


class CodebaseBranchesEndpoint(endpoint.EndpointMixin, unittest.TestCase):
    endpointClass = codebase_branches.CodebaseBranchesEndpoint
    resourceTypeClass = codebase_branches.CodebaseBranch

    @async_to_deferred
    async def setUp(self) -> None:  # type: ignore[override]
        await self.setUpEndpoint()
        await self.master.db.insert_test_data([
            fakedb.Project(id=7, name='fake_project7'),
            fakedb.Codebase(id=13, projectid=7, name='codebase1'),
            fakedb.CodebaseCommit(id=110, codebaseid=13),
            fakedb.CodebaseCommit(id=111, codebaseid=13),
            fakedb.CodebaseCommit(id=112, codebaseid=13),
            fakedb.CodebaseBranch(id=220, codebaseid=13, name='branch1', commitid=110),
            fakedb.CodebaseBranch(id=221, codebaseid=13, name='branch2', commitid=111),
            fakedb.CodebaseBranch(id=222, codebaseid=13, name='branch3', commitid=112),
        ])

    @parameterized.expand([
        ('no_filter', None, [220, 221, 222]),
        ('existing', 111, [221]),
        ('not_existing', 120, []),
    ])
    @async_to_deferred
    async def test_get(
        self, name: str, commitid_filter: int | None, expected_branchids: list[str]
    ) -> None:
        result_spec = None
        if commitid_filter is not None:
            result_spec = resultspec.OptimisedResultSpec(
                filters=[resultspec.Filter('commitid', 'eq', [commitid_filter])]
            )

        branches = await self.callGet(('codebases', 13, 'branches'), resultSpec=result_spec)

        for b in branches:
            self.validateData(b)

        self.assertEqual(sorted([b['branchid'] for b in branches]), expected_branchids)


class CodebaseBranchTests(interfaces.InterfaceTests, TestReactorMixin, unittest.TestCase):
    @async_to_deferred
    async def setUp(self) -> None:  # type: ignore[override]
        self.setup_test_reactor()
        self.master = await fakemaster.make_master(self, wantMq=True, wantDb=True, wantData=True)
        self.rtype = codebase_branches.CodebaseBranch(cast(BuildMaster, self.master))
        await self.master.db.insert_test_data([
            fakedb.Project(id=7, name='fake_project7'),
            fakedb.Codebase(id=13, projectid=7, name='codebase1'),
            fakedb.CodebaseCommit(id=110, codebaseid=13),
            fakedb.CodebaseCommit(id=111, codebaseid=13),
            fakedb.CodebaseBranch(id=220, codebaseid=13, name='branch1', commitid=110),
        ])

    def test_signature_update_branch(self) -> None:
        @self.assertArgSpecMatches(
            self.master.data.updates.update_branch,
            self.rtype.update_branch,
        )
        def update_branch(
            self: Any,
            *,
            codebaseid: int,
            name: str,
            commitid: int | None = None,
            last_timestamp: int,
        ) -> None:
            pass

    @async_to_deferred
    async def test_update_branch(self) -> None:
        await self.master.data.updates.update_branch(
            codebaseid=13,
            name='branch1',
            commitid=111,
            last_timestamp=87654321,
        )
        branch = await self.master.data.get(('branches', 220))
        self.assertEqual(
            branch,
            {
                'branchid': 220,
                'codebaseid': 13,
                'commitid': 111,
                'last_timestamp': 87654321,
                'name': 'branch1',
            },
        )



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

from buildbot.db.compression.protocol import CompressObjInterface
from buildbot.db.compression.protocol import CompressorInterface

try:
    import lz4.block

    HAS_LZ4 = True
except ImportError:
    HAS_LZ4 = False


class LZ4Compressor(CompressorInterface):
    name = "lz4"
    available = HAS_LZ4

    @staticmethod
    def dumps(data: bytes) -> bytes:
        return lz4.block.compress(data)

    @staticmethod
    def read(data: bytes) -> bytes:
        return lz4.block.decompress(data)

    # LZ4.block does not have a compress object,
    # still implement the interface for compatibility
    class CompressObj(CompressObjInterface):
        def __init__(self) -> None:
            self._buffer: list[bytes] = []

        def compress(self, data: bytes) -> bytes:
            self._buffer.append(data)
            return b''

        def flush(self) -> bytes:
            compressed_buffer = LZ4Compressor.dumps(b''.join(self._buffer))
            self._buffer = []
            return compressed_buffer
